/* Copyright 2006 Ikeda Yasuyuki
** 
** This file is part of ipenv.
** 
** ipenv is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** any later version.
** 
** ipenv is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with ipenv; if not, write to the Free Software
** Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
** 
*/

/**
 * mod_ipenv.c -- an apache2 module to set/unset environment variables
 *                base on the ip address or the host name of the remote hosts.
 * 
 * To use ipenv, make ipenv.so with apxs.
 * 
 *   # apxs -c -i mod_ipenv.c
 * 
 * Then activate it in Apache2's httpd.conf file for instance
 * 
 *   # cat httpd.conf
 *   ...
 *   LoadModule ipenv_module modules/mod_ipenv.so
 *   ...
 * 
 * You can control environment variables with SetEnvIp directives.
 * 
 *   # cat access/localnet.conf
 *   127.0.0.1
 *   192.168.0.0/24
 * 
 *   # cat httpd.conf
 *   ...
 *   <Location /local>
 *     SetEnvIp access/localnet.conf ADMIN
 *     Order deny, allow
 *     deny from all
 *     allow from ADMIN
 *   </Location>
 * 
 * Then after restarting Apache via
 * 
 *    $ apachectl restart
 * 
 * @author Ikeda Yasuyuki
 * @version 1.0.1
 */

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_core.h"
#include "ap_config.h"
#include "apr_strings.h"

#include "mod_ipenv.h"

#ifdef DEBUG
#include<stdio.h>
#endif /*DEBUG*/

/*pre declaration*/
module AP_MODULE_DECLARE_DATA ipenv_module;



/**
 * called for configurations for each directory
 * (includes root directory == server global).
 * create a new directory configutaion instance.
 *
 * @param	p	apr pool object.
 * @param	path	path to the directory.
 * 
 * @return	ipenv directory configuration object.
 */
static void* ipenv_create_dir_config(apr_pool_t* p, char* path){
	ipenv_dir_conf* c;
	
	c = (ipenv_dir_conf*)apr_pcalloc(p, sizeof(ipenv_dir_conf));
	
#ifdef DEBUG
	fprintf(stderr, "create_dir: %s\n", (path == NULL)?"(none)":path);
#endif /*DEBUG*/
	
	if(path == NULL){
	/*root directory == server global*/
		c->path = NULL;
		c->entry_list = apr_array_make(p, IPENV_INIT_ENTRYSIZE, sizeof(ipenv_entry));
	}else{
		c->path = apr_pstrdup(p, path);
		c->entry_list = apr_array_make(p, IPENV_INIT_ENTRYSIZE, sizeof(ipenv_entry));
	}
	
	return (void*)c;
}

/**
 * called for configurations for each directories.
 * We must merge the configuration of the directory
 * and the base one (i.e. that of the parent directory).
 * and MUST allocate new configuration object.
 */
static void* ipenv_merge_dir_config(apr_pool_t* p, void* base_conf, void* new_conf){
	ipenv_dir_conf* base_c;
	ipenv_dir_conf* new_c;
	ipenv_dir_conf* c;
	
	base_c = (ipenv_dir_conf*)base_conf;
	new_c = (ipenv_dir_conf*)new_conf;
	c = (ipenv_dir_conf*)apr_pcalloc(p, sizeof(ipenv_dir_conf));
	
#ifdef DEBUG
	fprintf(stderr, "merge_dir: %s %s\n"
		, (base_c->path == NULL)?"(none)":base_c->path
		, (new_c->path == NULL)?"(none)":new_c->path
	);
#endif /*DEBUG*/
	
	c->path = new_c->path;
	c->entry_list = apr_array_append(p, base_c->entry_list, new_c->entry_list);
		/*Base ones should be evaluated first.*/
	
	return (void*)c;
}

/**
 * convert string to ipenv_remote.
 * 
 * @param	r	ipenv_remote object.
 * @param	s	string
 * @param	p_err	place to store error string.
 *
 * @return	0 for success, non-0 otherwise.
 */
static int ipenv_read_entry(apr_pool_t* p, ipenv_remote* r, char* s, const char** p_err){
	char* t;
	apr_status_t rv;
	char ebuffer[EBUFFER_SIZE];
	
	if((t = strchr(s, '/')) != NULL){
		*t++ = '\0';
		rv = apr_ipsubnet_create(&(r->x.ip), s, t, p);
		if(APR_STATUS_IS_EINVAL(rv)){
			r->type = IPENV_TYPE_FAIL;
			*p_err = apr_psprintf(p, "bad ip format");
			return 1;
		}else if (rv != APR_SUCCESS) {
			r->type = IPENV_TYPE_FAIL;
			apr_strerror(rv, ebuffer, sizeof(ebuffer));
			*p_err = apr_pstrdup(p, ebuffer);
			return 2;
		}
		r->type = IPENV_TYPE_IP;
	}else if(!APR_STATUS_IS_EINVAL(rv = apr_ipsubnet_create(&(r->x.ip), s, NULL, p))){
		if(rv != APR_SUCCESS){
			r->type = IPENV_TYPE_FAIL;
			apr_strerror(rv, ebuffer, sizeof(ebuffer));
			*p_err = apr_pstrdup(p, ebuffer);
			return 3;
		}
		r->type = IPENV_TYPE_IP;
	}else if(s[0] == '.'){
		r->type = IPENV_TYPE_DOMAIN;
		r->x.domain = apr_pstrdup(p, s);
	}else{
		r->type = IPENV_TYPE_HOST;
		r->x.host = apr_pstrdup(p, s);
	}
	
	return 0;
}


/**
 * open file, and read lines into remote_entry object.
 * 
 * @param	p	apr pool object.
 * @param	filename
 * @param	p_err	place to store error string.
 *
 * @return	apr_array_header_t object of remote_entry. NULL for errors.
 */
static apr_array_header_t* ipenv_read_file(apr_pool_t* p, const char* filename, const char** p_err){
	apr_array_header_t* remote_list;
	ipenv_remote* r;
	
	apr_file_t* fp = NULL;
	int line_number;
	char buffer[BUFFER_SIZE];
	char* s;
	const char* cursor;
	
#ifdef DEBUG
	fprintf(stderr, "\topen: %s\n", filename);
#endif /*DEBUG*/
	
	
	if(apr_file_open(&fp, filename, APR_READ, APR_OS_DEFAULT, p) != APR_SUCCESS){
		*p_err = apr_pstrcat(p, "fail to open: ", filename, NULL);
		return NULL;
	}
	
	remote_list = apr_array_make(p, 10, sizeof(ipenv_remote));
	
	
	line_number = 0;
	while(remote_list != NULL && apr_file_gets(buffer, sizeof(buffer), fp) == APR_SUCCESS){
		++line_number;
		cursor = buffer;
		while(*cursor != '\0'){
			s = ap_getword_white(p, &cursor);
			if(s[0] == '\0')continue;
			if(s[0] == '#')break;
			
			r = (ipenv_remote*)apr_array_push(remote_list);
			
			if(ipenv_read_entry(p, r, s, p_err)){
				*p_err = apr_pstrcat(p, "fail to parse in %s at line %d: %s"
					, filename, line_number, *p_err, NULL
				);
				remote_list = NULL;
				break;
			}
		}
	}
	
    apr_file_close(fp);
	
	return remote_list;
}

/**
 * SetEnvIp filename [!]envname
 * 
 * @param	cmd	apache command parameter object.
 * @param	dconf	(ipenv) directory configuration object.
 * @param	filename
 * @param	envname	name of the environment variable, maybe prefixed with '!'
 * 
 * @return	NULL for no error, error string otherwise.
 */
static const char* ipenv_set_env_ip(cmd_parms* cmd, void* dconf, const char* filename, const char* envname){
	ipenv_dir_conf* c;
	ipenv_entry* e;
	const char* err = NULL;
	
	c = (ipenv_dir_conf*)dconf;	/*is this always not NULL?*/
	e = (ipenv_entry*)apr_array_push(c->entry_list);
	
	filename = ap_server_root_relative(cmd->pool, filename);
	
#ifdef DEBUG
	fprintf(stderr, "SetEnvIp %s %s on %s\n", filename, envname, (c->path == NULL)?"(root)":c->path);
#endif /*DEBUG*/
	
	/*load entries from file.*/
	if((e->remote_list = ipenv_read_file(cmd->pool, filename, &err)) == NULL)
		return apr_pstrcat(cmd->pool, "SetEnvIp: ", err, NULL);
	
	e->filename = apr_pstrdup(cmd->pool, filename);
	
	if(envname[0] == '!'){
		e->setenv = IPENV_UNSET;
		++envname;
	}else{
		e->setenv = IPENV_SET;
	}
	e->envname = apr_pstrdup(cmd->pool, envname);
	
	return NULL;
}

static const command_rec ipenv_cmds[] = {
	AP_INIT_TAKE2("SetEnvIp"
		,ipenv_set_env_ip
		,NULL
		, OR_FILEINFO
		, "SetEnvIp filename [!]envname"
	)
	, {NULL}
};


/**
 * test if a hostname matches a domain.
 * simply tests if the domain is the suffix of the hostname.
 * 
 * @param	hostname
 * @param	domain	domain name prefixed with ".".
 * 
 * @return	non-0 if match, 0 otherwise.
 */
static int ipenv_match_domain(const char* hostname, const char* domain){
	int hl;
	int dl;
	
	hl = strlen(hostname);
	dl = strlen(domain);
	
	return (
		((hl - dl) > 0)
		&& (strcasecmp(hostname + (hl - dl), domain) == 0)
	);
}

/**
 * check if a request matches a ipenv entry(that is, each SetEnvIp directive)
 * 
 * @param	r	apache2 request object.
 * @param	e	ipenv_entry object.
 * @return	non-0 if matched, 0 otherwise.
 */
static int ipenv_check(request_rec* r, const ipenv_entry* e){
    ipenv_resolve resolve;
    const char* remotehost;
	int remotehost_is_ip;
	
	const ipenv_remote* ipe;
	
	int i;
	
#ifdef DEBUG
	fprintf(stderr, "\tchecking againt %s\n", e->filename);
#endif /*DEBUG*/
	
	resolve = IPENV_NOTRESOLVED;
	remotehost = NULL;
	
	for(i = 0; i < e->remote_list->nelts; ++i){
		ipe = &(((const ipenv_remote*)e->remote_list->elts)[i]);
		
		if((ipe->type == IPENV_TYPE_HOST || ipe->type == IPENV_TYPE_DOMAIN)
			&& resolve == IPENV_NOTRESOLVED
		){
			remotehost = ap_get_remote_host(
				r->connection
				, r->per_dir_config
				, REMOTE_DOUBLE_REV
				, &remotehost_is_ip
			);
			resolve = ((remotehost != NULL) && !remotehost_is_ip)
				?IPENV_RESOLVED
				:IPENV_RESOLVEFAILED;
		}
		
		switch (ipe->type) {
		case IPENV_TYPE_IP:
			if(apr_ipsubnet_test(ipe->x.ip, r->connection->remote_addr))return 1;
			break;
		case IPENV_TYPE_HOST:
			if(resolve == IPENV_RESOLVED
				&& strcasecmp(remotehost, ipe->x.host) == 0
			)return 1;
			break;
		case IPENV_TYPE_DOMAIN:
			if(resolve == IPENV_RESOLVED
				&& ipenv_match_domain(remotehost, ipe->x.domain)
			)return 1;
			break;
		default:
			break;
		}
	}
	return 0;
}

/**
 * Do main process. hook this wherever you like.
 * test all ip/host entries,
 * and set/unset an environment variable when matched.
 * 
 * @param	r	apache2 request object.
 * 
 * @return	always return DECLINES
 */
static int ipenv_process(request_rec* r){
	int i;
	const ipenv_dir_conf* c;
	const ipenv_entry* e;
	
	c = (const ipenv_dir_conf*)ap_get_module_config(r->per_dir_config, &ipenv_module);
	
#ifdef DEBUG
	fprintf(stderr, "ipenv invoked\n");
#endif /*DEBUG*/
	
	/* if no directory configuration is available,
		return immediately.*/
	if(c == NULL)
		return DECLINED;
	
#ifdef DEBUG
	fprintf(stderr, "ipenv in %s\n", (c->path == NULL)?"(root)":c->path);
#endif /*DEBUG*/
	
	for (i = 0; i < c->entry_list->nelts; ++i){
		e = &(((const ipenv_entry*)c->entry_list->elts)[i]);
		
		if(!ipenv_check(r, e))continue;
		
#ifdef DEBUG
		fprintf(stderr, "\tmatched\n");
#endif /*DEBUG*/
		
		switch(e->setenv){
		case IPENV_SET:
			apr_table_setn(r->subprocess_env, e->envname, "1");
#ifdef DEBUG
			fprintf(stderr, "\tset %s\n", e->envname);
#endif /*DEBUG*/
			break;
		case IPENV_UNSET:
			apr_table_unset(r->subprocess_env, e->envname);
#ifdef DEBUG
			fprintf(stderr, "\tunset %s\n", e->envname);
#endif /*DEBUG*/
			break;
		}
	}
	
	return DECLINED;
}

static const char* const post_modules[] = { "mod_setenvif.c", NULL };

static void ipenv_register_hooks(apr_pool_t* p){
	ap_hook_header_parser(ipenv_process, NULL, post_modules, APR_HOOK_MIDDLE);
		/*This seems the best hook point to make ipenv useful.*/
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA ipenv_module = {
	STANDARD20_MODULE_STUFF
	, ipenv_create_dir_config	/* create per-dir	config structures */
	, ipenv_merge_dir_config	/* merge  per-dir	config structures */
	, NULL				/* create per-server config structures */
	, NULL				/* merge  per-server config structures */
	, ipenv_cmds	/* table of config file commands */
	, ipenv_register_hooks	/* register hooks  */
};

